@@ -27,8 +27,6 @@ import subprocess as sp
 import sys
 import types
 
-import _pytest
-
 from pyls_livepy import __version__ as plr_version
 from pyls_livepy.utils import (
     _mark_err,
@@ -38,16 +36,18 @@ from pyls_livepy.utils import (
 )
 
 
-MODULE_UNDER_TEST = None
+class PytestException(BaseException):
+    """Signal exceptions specific to the pytest case runner."""
+    pass
 
 
-def _get_stdin():
-    return "".join([ln for ln in sys.stdin])
+class PytestCaseRunnerSuccess(BaseException):
+    """Signal successful case run.
 
-
-def _put_stdout(out):
-    sys.stdout.write(out)
-    return
+    A hack to avoid the failure message, caused by the report output.
+    """
+    # TODO: find a proper way to ignore a passing result
+    pass
 
 
 class CapDocTestRunner(doctest.DocTestRunner):
@@ -85,8 +85,25 @@ class CapDocTestRunner(doctest.DocTestRunner):
         return
 
 
-class PytestException(BaseException):
-    pass
+class PytestCaseRunner:
+    _mut = None
+    mname = __name__
+
+    def __call__(self, argstr):
+        return _run_pytest_case(argstr, self._mut)
+
+    def set_mut(self, module):
+        self._mut = module
+        return
+
+
+def _get_stdin():
+    return "".join([ln for ln in sys.stdin])
+
+
+def _put_stdout(out):
+    sys.stdout.write(out)
+    return
 
 
 def capout(hold=None):
@@ -105,16 +122,16 @@ def capout(hold=None):
     return inner
 
 
-def run_doctests(src, which=None):
+def run_doctests(src, mut_path, which=None):
     """Run the doctests in a document."""
-    # TODO: move to runner module
     run_env = sp.run(
         ["bash", "-c", "which pyls-livepy-runner"],
         capture_output=True,
         text=True,
     ).stdout.strip()
     logger.info(f'Runner at "{run_env}"')
-    mod = types.ModuleType("mod")  # TODO: get name from document
+    mod = types.ModuleType("rdt_mut")  # TODO: get name from document
+    mod.__file__ = mut_path
     w_path = "\n  ".join(sys.path)
     logger.info(f"Working PATH:\n  {w_path}\n")
 
@@ -124,8 +141,6 @@ def run_doctests(src, which=None):
     except Exception as e:
         logger.exception(f"Exception while exec'ing MUT: {repr(e)}")
         return {}
-    global MODULE_UNDER_TEST
-    MODULE_UNDER_TEST = mod
     doctests = doctest.DocTestFinder().find(mod)
     runner = CapDocTestRunner()
     results = {}
@@ -134,9 +149,17 @@ def run_doctests(src, which=None):
         # TODO: allow filtering by 'which'
         if not dt.examples:
             continue
+        _maybe_update(dt, mod)
         co = capout()
         res = runner.run(dt, out=co)._asdict()
-        bad = [tr for tr in co() if tr["type"] in ["failure", "exception"]]
+        bad = [
+            tr for tr in co()
+            if tr["type"] in ["failure", "exception"]
+            and (
+                not isinstance(tr["exc_info"][1], PytestCaseRunnerSuccess)
+                if "exc_info" in tr else True
+            )
+        ]
         res["out"] = bad
         results[dt.name] = res
     return results
@@ -189,12 +212,13 @@ def run():
     src = run_data["source"]
     plp_version = run_data["plp_version"]
     log_level = run_data["log_level"]
+    mut_path = run_data["mut_path"]
     set_log_level(log_level, logger)
     logger.debug(f'sys.prefix = "{sys.prefix}"')
     logger.info(f'Runner active at "{Path(__file__).parent}"')
 
     if plp_version == plr_version:
-        markers = create_markers(run_doctests(src))
+        markers = create_markers(run_doctests(src, mut_path))
 
     else:
         err = "version mismatch error; "
@@ -207,35 +231,98 @@ def run():
 
 def _perform_collect_wrapper(self, args=None, genitems=True):
     """Modify wrapped's return value."""
-    items = self._wrapped_perform_collect(args, genitems)
-    global MODULE_UNDER_TEST
-    mod = MODULE_UNDER_TEST
-    # NOTE: assumes MUT name is MUT tester name with test prepended
-    # TODO: should prob get the name from sys.modules
-    mutt_name = items[0].module.__name__
-    mut_name = mutt_name.partition("_")[2]
-    mod.__name__ = getattr(items[0].module, mut_name).__name__
-    mod.__file__ = getattr(items[0].module, mut_name).__file__
-    setattr(items[0].module, mut_name, mod)
-    items[0]._reportscap = self._reportscap
+
+    try:
+        items = self._wrapped_perform_collect(args, genitems)
+        mut = self._mut
+        assert mut is not None, "MUT transfer failed"
+        # NOTE: assumes MUT name is MUT tester name with test prepended
+        # TODO: should prob get the name from sys.modules or so
+        mutt_name = items[0].module.__name__
+
+        mut_name = mutt_name.partition("_")[2]
+        mut.__name__ = getattr(items[0].module, mut_name).__name__
+        mut.__file__ = getattr(items[0].module, mut_name).__file__
+        setattr(items[0].module, mut_name, mut)
+        items[0]._reportscap = self._reportscap
+
+    except Exception as e:
+        logger.exception(repr(e))
+        raise
     return items
 
 
 def _runtestprotocol_wrapper(item, log=True, nextitem=None):
     """Capture test results from runtestprotocol."""
-    reports = _pytest.runner._wrapped_runtestprotocol(item, log, nextitem)
+    import _pytest.runner
 
+    reports = _pytest.runner._wrapped_runtestprotocol(item, log, nextitem)
     item._reportscap(reports)
     return reports
 
 
-def run_pytest_case(argstr):
+def _maybe_update(doctest, module):
+    """Maybe apply changes to the doctest before running it."""
+    upd = False
+
+    for o_name, obj in doctest.globs.items():
+
+        if getattr(obj, "name", "") == "pyls_livepy.run_pytest_case":
+            import ast
+
+            obj.set_mut(module)
+            upd = True
+            rpc_imp = ""
+
+            for ex in doctest.examples:
+                pex = ast.parse(ex.source)
+
+                for node in ast.walk(pex):
+                    imp = isinstance(node, ast.Import)
+                    impf = isinstance(node, ast.ImportFrom)
+                    impf = impf and node.module == "pyls_livepy"
+                    rpc_call = isinstance(node, ast.Call)
+                    rpc_call = rpc_call and getattr(node.func, "id") == rpc_imp
+
+                    if rpc_call:
+                        ex.want = ""
+                        break
+
+                    if imp or impf and any([
+                            n.name in [
+                                "run_pytest_case",
+                                "pyls_livepy.run_pytest_case",
+                            ]
+                            for n in node.names
+                    ]):
+                        # get the actual name
+                        rpc_imp = [
+                            n.asname if n.asname else n.name
+                            for n in node.names if n.name in [
+                                    "run_pytest_case",
+                                    "pyls_livepy.run_pytest_case",
+                            ]
+                        ][0]
+                        logger.debug(f"RPC = {rpc_imp}")
+                        break
+    return upd
+
+
+def _run_pytest_case(argstr, mut):
     """Run the pytest case specified by argstr.
 
     Patches pytest to insert the updated module and collect the test result.
 
     >>> run_pytest_case("-k run_doctests --no-cov")
     """
+    try:
+        import _pytest
+        # NOTE: hack to prevent AttributeError
+        import _pytest.main as pm
+        pm
+
+    except ModuleNotFoundError:
+        raise PytestException("pytest must be installed to use this function")
     Session = _pytest.main.Session
     runner = _pytest.runner
     reportscap = capout()
@@ -243,6 +330,7 @@ def run_pytest_case(argstr):
 
     Session._wrapped_perform_collect = Session.perform_collect
     Session.perform_collect = _perform_collect_wrapper
+    Session._mut = mut
     runner._wrapped_runtestprotocol = runner.runtestprotocol
     runner.runtestprotocol = _runtestprotocol_wrapper
     args = shlex.split(argstr)
@@ -251,7 +339,11 @@ def run_pytest_case(argstr):
     del Session._wrapped_perform_collect
     runner.runtestprotocol = runner._wrapped_runtestprotocol
     del runner._wrapped_runtestprotocol
-    reports = reportscap()[0]
+    captured = reportscap()
+
+    if not captured:
+        raise PytestException("no reports captured")
+    reports = captured[0]
 
     for rep in reports:
         if rep.longrepr:
@@ -265,4 +357,8 @@ def run_pytest_case(argstr):
             except Exception as e:
                 logger.error(repr(e))
                 raise LivepyError(str(e))
-    return
+    raise PytestCaseRunnerSuccess()  # NOTE: signal test case passed
+
+
+run_pytest_case = PytestCaseRunner()
+run_pytest_case.name = "pyls_livepy.run_pytest_case"
